/*
DUMP1 -- Dump a file in hex and ASCII
    Copyright (C) 2003  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with Visual Studio 6 or later.  Requires comdlg32.lib odbc32.lib odbccp32.lib winmm.lib
	cl xfer32.c /link comdlg32.lib odbc32.lib odbccp32.lib winmm.lib

*/


#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#ifdef WIN32
#include <io.h>
#define OPENFLAG O_BINARY | O_RDONLY
#else
#define OPENFLAG O_RDONLY
#include <sys/types.h>
#include <unistd.h>
#define _read read
#define _open open
#endif

unsigned char buffer[4096];

char hextable[][3] = {
"00","01","02","03","04","05","06","07","08","09","0A","0B","0C","0D","0E","0F",
"10","11","12","13","14","15","16","17","18","19","1A","1B","1C","1D","1E","1F",
"20","21","22","23","24","25","26","27","28","29","2A","2B","2C","2D","2E","2F",
"30","31","32","33","34","35","36","37","38","39","3A","3B","3C","3D","3E","3F",
"40","41","42","43","44","45","46","47","48","49","4A","4B","4C","4D","4E","4F",
"50","51","52","53","54","55","56","57","58","59","5A","5B","5C","5D","5E","5F",
"60","61","62","63","64","65","66","67","68","69","6A","6B","6C","6D","6E","6F",
"70","71","72","73","74","75","76","77","78","79","7A","7B","7C","7D","7E","7F",
"80","81","82","83","84","85","86","87","88","89","8A","8B","8C","8D","8E","8F",
"90","91","92","93","94","95","96","97","98","99","9A","9B","9C","9D","9E","9F",
"A0","A1","A2","A3","A4","A5","A6","A7","A8","A9","AA","AB","AC","AD","AE","AF",
"B0","B1","B2","B3","B4","B5","B6","B7","B8","B9","BA","BB","BC","BD","BE","BF",
"C0","C1","C2","C3","C4","C5","C6","C7","C8","C9","CA","CB","CC","CD","CE","CF",
"D0","D1","D2","D3","D4","D5","D6","D7","D8","D9","DA","DB","DC","DD","DE","DF",
"E0","E1","E2","E3","E4","E5","E6","E7","E8","E9","EA","EB","EC","ED","EE","EF",
"F0","F1","F2","F3","F4","F5","F6","F7","F8","F9","FA","FB","FC","FD","FE","FF"};

char *memsearch(char *s,size_t ls, char *t, size_t lt) {
	size_t i;
	int match;
	// search memory starting at location s for the 
	// pattern stored at t;
	// ls is the size of s
	// lt is the size of t
	
	for (i=0; i <= (ls - lt); ++i) {
		match =  memcmp(s+i,t,lt);
		if (match == 0)
			// found a match
			return s+i;
	}
	return NULL;
}

long xtol(char *s) {
	long result;
	int factor;
	for (result=0;*s;++s) {
		switch(*s = toupper(*s)) {
			// values 10-15
			case 'A':
			case 'B':
			case 'C':
			case 'D':
			case 'E':
			case 'F':
				factor = 10 + *s - 'A';
				break;
			// values 0 - 9
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
        		factor = *s - '0';
				break;
			// linefeed means end of string
			case '\n':
				return result;
			// invalid value, return 0
			default:
        		return 0l;
		}
    	result = (result << 4) + factor;
	}
	return result;
}

long ctol(char *s) {
	long result;
    for (result=0;*s;++s) {
	    if (*s >= '0' && *s <= '9')
        	result = 10*result + *s - '0';
        else
        	return 0l;
	}
	return result;
}

long new_offset(char *s, int relative, int previous_offset) {
	long start_offset;
	// An offset was entered either in hex or decimal
	start_offset = toupper(*(s+1)) == 'X' ? xtol(s+2) : atol(s);
	// Now apply either as absolute or relative
	switch(relative) {
	case -1:
		// relative offset back
		start_offset = previous_offset - start_offset;
		if (start_offset < 0)
			start_offset = 0;
		break;
	case 1:
		// relative offset forward
		start_offset += previous_offset;					
		break;
	case 0:
		// absolute offset, nothing more to do
		break;
	}
	return start_offset;
}


int isrelative(char s) {
// Check to see if offset is absolute (no sign) or relative (+ forward, - backward)
int relative;
switch(s) {
	case '+':
		// Relative going forward
		relative = 1;
		break;
	case '-':
		// relative going backward
		relative = -1;
		break;
	default:
		// absolute
		relative = 0;
	}
    return relative;
}

// update the start_offset based on command
// return -2 to quit, -1 continuous, else 0
int updateoffset(char *s,long *start_offset, int relative, long previous_offset) {
	switch (toupper(*s)) {
	case 'S':
		return -3;		// search
	case 'Q':
		return(-2);	// quit
	case 'C':
		return -1;		// continuous
		break;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		*start_offset = new_offset(s,relative,previous_offset);
		break;
	default:
		// if + entered, move down 1 line.  If - entered, move up 1 line
		switch(relative) {
		case 1:  // down 1 line
			*start_offset = previous_offset + 16;
			break;
		case -1: //up 1 line
			*start_offset = previous_offset - 16;
			break;
		case 0: //do nothing
			break;
		}
		break;
	}
	return 0;
}

char ebcdic[256], reverse_ebcdic[256]; // used to translate between ebcdic and ascii
unsigned char searchstr[1024], strtype=' ';
size_t searchlen;
long searchFile(unsigned file, long previous_offset) {
	unsigned char buffer[4096],ignorecase[5], *found, *b;
	size_t numin,rdlen,buflen;
	char again = 0;
	// prompt for search str
	// position to previous_offset
	// scan for the search str
	// return offset of found string or current_pos if not found
	//
	printf("Search for(H:data-hex, E:data-EBCDIC, A:data|data-ASCII,RETURN repeat):\n");
	fgets((char *)buffer,sizeof(buffer),stdin);
	buflen = strlen((char *)buffer) - 1;
	buffer[buflen] = 0; //remove terminating /n

	if (buflen == 0) 
		// repeat last srch
		if (strtype==' ')
			// nothing to repeat
			return (previous_offset);
		else {
			// look for a new occurrence
			previous_offset += searchlen;
			again = 1;
		}
	else
		strtype = 'A';  // set default type

	if (buflen > 2) {
		if (buffer[1] == ':') {
			strtype = toupper((int) buffer[0]);
			b = buffer+2;
		}
		else
			b = buffer;
	}
		
	// Now process depending on type	
	if (!again)

		switch(strtype) {
		case 'A':
		case 'E':
			searchlen = strlen((char *)b);
			if (searchlen > sizeof(searchstr)) {
				printf("String too long.  Must not exceed %d characters.\n",sizeof(searchstr));
				return  0;
			}
//			printf("Ignore case (Y or N)? ");
//			fgets(ignorecase,sizeof(ignorecase),stdin);
			if (strtype == 'A')
				strcpy((char *)searchstr,(char *)b);
			else {
				while(*b) 
					searchstr[searchlen++] = reverse_ebcdic[*b++];
			}
			//copy from buffer to searchstr translating from EBCDIC
			break;
		case 'H':
			// data entered as xx xx xx, translate every two characters to 1 byte
			// can enter an ascii string enclosed in " , use \ as escape character to enter a "
			// and store in search str
			searchlen = 0;
			searchstr[0] = 0; //set searchstr to the empty string
			found = (unsigned char *)strtok((char *)b, " ");
			while(found != NULL) {
				// copy info for buffer to searchstr
				if (*found == '"') {
					if (found[strlen((char *)found)-1] == '"') {
						found[strlen((char *)found)-1] = 0;
					}
					// Entering a string, start after " stop at end
					++found;
					while (*found) {
						searchstr[searchlen++] = *found++;
					}
				}
				else 	
					if (strlen((char *)found) < 2 || strchr("0123456789ABCDEFabcdef",*found) == NULL || strchr("0123456789ABCDEFabcdef",*(found+1)) == NULL) {
						// Not a hex string
						fprintf(stderr,"\n%s -- Not hexadecimal string.",found);
						return previous_offset;
					}
				searchstr[searchlen++] = xtol((char *)found);
				found = (unsigned char *)strtok((char *)NULL," ");
			}
			searchstr[searchlen] = 0;
//			searchlen--;  // set searchlen to actual length of buffer
			break;
		default:
			fprintf(stderr,"%c -- Invalid string type.\n",strtype);
			return previous_offset;
	}

	// Now do search.  Read in 1 blk,  scan for match up to position blocksize - len(searchstr);
	// If no match, copy rest of buffer to start of buffer, fill in rest of buffer, continue to eof
	// If match, return offset; else return -1
	
	rdlen = sizeof(buffer);
	lseek(file,previous_offset,SEEK_SET); // position file to starting position
	do {
		// Read in data 1 block at a time
		numin = _read(file,buffer+sizeof(buffer)-rdlen,rdlen);
		found = (unsigned char *)memsearch((char *)buffer,numin,(char *)searchstr,searchlen);
		if (found != NULL) {
			// found a match in buffer
			return (lseek(file,0,SEEK_CUR) - sizeof(buffer) + (found - buffer));
		}
		// no match
		memcpy(buffer,buffer+sizeof(buffer)-searchlen,searchlen);
		// from now on, readin sizeof(buffer) - searchlen bytes
		rdlen = sizeof(buffer) - searchlen;
	} while(numin == rdlen);		
	// Not found, position to previous_offset
	fprintf(stderr,"Not Found.\n");
	return previous_offset;
}



int main(int argc,char *argv[]) {
	unsigned file,linekt;
	int numin, cfrom, relative;
	size_t ln,i;
	long start_offset,end_offset,tot,previous_offset = 0;
	char c,eflag,cflag,*finm,*start,*end,str[80],*s, *b;
	char current_line[80], previous_line[80], dotdotdot = 0;
	FILE *hlp = stderr;

	if (argc < 2) {
		fprintf(hlp,"DUMP1 Version 1.01, %s\n",__DATE__);
		fprintf(hlp,"\ncopyright 2003, Norm London, London Computing.\nnorm_london@londoncomputing.com\n\n");
        	fprintf(hlp,"This program is free software; you can redistribute it and/or modify\n");
		fprintf(hlp,"it under the terms of the GNU General Public License as published by\n");
		fprintf(hlp,"the Free Software Foundation; either version 2 of the License, or\n");
		fprintf(hlp,"(at your option) any later version.\n\n");
		fprintf(hlp,"This program is distributed in the hope that it will be useful,\n");
		fprintf(hlp,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
		fprintf(hlp,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
		fprintf(hlp,"GNU General Public License for more details.\n\n");
		fprintf(hlp,"You should have received a copy of the GNU General Public License\n");
		fprintf(hlp,"along with this program; if not, write to the Free Software\n");
		fprintf(hlp,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");

		fprintf(stderr,"DUMP1 Copyright (c) London Computing 1988-2003.\n");
		fprintf(stderr,"Usage is:  DUMP1 filename [start offset] [end offset] [-Ebcdic] [-Continuous]\n");
		fprintf(stderr,"Dumps a file starting at start offset (default 0) to end offset (default EOF).\n");
		fprintf(stderr,"Dumps 20 lines (320 bytes) at a time in hexidecimal and character notation.\n");
		fprintf(stderr,"-E -- Assume text is using EBCDIC, not ASCII.\n");
		fprintf(stderr,"-C -- Dump continuously.  Default is 1 page at a time.\n\n");
		fprintf(stderr,"After each page is displayed, you can press RETURN to see the next page,\n");
		fprintf(stderr,"Q to Quit, C to display continuously or enter a new starting offset\n");
		fprintf(stderr,"either in decimal (ddd) or in hex (0xhhh).  If the value starts with \n");
		fprintf(stderr,"either + or -, it is realtive to the current position.\n");
		return(1);
	}
	finm = NULL;
	start=NULL;
	end = NULL;
	start_offset = end_offset = 0;
	tot=0;
	eflag = 0;
	cflag = 0;
	previous_line[0] = 0;

	for (i=1;i < argc;++i) {
    	    if (argv[i][0] == '-') {
    		switch(toupper(argv[i][1])) {
    		case 'E': eflag = 1; break;
    		case 'C': cflag = 1; break;
    		default:  fprintf(stderr,"%ls <=== Flag ignored!",argv[i]);
    	    }
    	}
    	else if (finm == NULL) {
			if ((file=_open(argv[i],OPENFLAG)) == -1) {
				fprintf(stderr,"Cannot Open %;s.",argv[1]);
				return(1);
			}
			finm = argv[i];
		}
		else if (start == NULL) {
			start = argv[i];
			if (*(start+1)=='x' || *(start+1) == 'X')
				start_offset = xtol(start+2);
			else
				start_offset = ctol(start);
		}
		else if (end == NULL) {
			end = argv[i];
			if (*(end+1)=='x' || *(end+1)=='X')
				end_offset = xtol(start+2);
			else
				end_offset = ctol(start);
		}
    }
//	if (eflag) {
		for (i=0;i<256;ebcdic[i++]=0) ; /* build translate table, printable only*/
		ebcdic[64]=' ';
		strncpy(ebcdic+75,".<(+|&",6);
		strncpy(ebcdic+90,"!$*);^-/",8);
		strncpy(ebcdic+106,"|,%_>?",6);
		strncpy(ebcdic+121,"`:#@\'=\"",7);
		strcpy(ebcdic+129,"abcdefghi.......jklmnopqr.......~stuvwxyz");
		strcpy(ebcdic+192,"{ABCDEFGHI......}JKLMNOPQR......\\.STUVWXYZ......0123456789|");

		//build reverse translate table
		for (i=0;i<256;++i) {
			reverse_ebcdic[i] = 0;
			if (ebcdic[i])
				reverse_ebcdic[ebcdic[i]] = i;
		}
//	}
	/*	read and dump 4k at a time	*/
	if (end_offset == 0)
		end_offset = 0x7fffffff;

	if (start_offset) {
		lseek(file,start_offset,0);
		tot = start_offset;
	}
	
	linekt = 0;
	do {
		numin=_read(file,buffer,4096);
		if (numin == -1) {
			fprintf(stderr,"Cannot Read %s.",argv[1]);
			return(1);
		}
		cfrom=0;
//		linekt=0;
		while (cfrom < numin) {
			relative = 0;
			if (linekt > 20 && cflag == 0) {
				start_offset = -1;  // indicate no offset change
				printf("\nMore...[RETURN-more,Q-quit,C-Continuous,S-Search,#####--new offset]: ");
				if (fgets(str,sizeof(str),stdin) == NULL)
					return(0);
				for (s=str;*s==' ';++s); /* ignore leading spaces */
				for (i=strlen(s);--i && *(s+i) == ' ';); /* and trailing spaces */
				previous_line[8] = 0; // force a no match
				relative = isrelative(*s); // is request relative or absolute ?
				if (relative)
					// move past sign
					++s;
				// Check to see what was entered
				cflag = updateoffset(s, &start_offset, relative, previous_offset);
				switch(cflag) {
					case -3: //search
						start_offset = searchFile(file,previous_offset);
						cflag = 0;
						break;
					case -2: // quit
						return 0;
					case -1: // continuous
						cflag = 1;
						break;
					default:
						break;
				}


				linekt = 0;
				if (start_offset >= 0) {
				// position file to new offset
					end_offset = 0x7fffffff;
					lseek (file,start_offset,SEEK_SET);
					tot = start_offset;
					cfrom = 0;
					numin=_read(file,buffer,4096);
					if (numin == -1) {
						fprintf(stderr,"Cannot Read %s.",argv[1]);
						return(1);
					}
					if (numin == 0)
						return(0);
				}
			}

			if (linekt == 0)
				previous_offset = cfrom+tot;
			/*	print the offset in hex	*/
			sprintf(current_line,"%8x",(long)(cfrom+tot)&0xffffffff);
			for (b=current_line;*b==' ';*b++ = '0');
	
			/*	print 16 bytes in hex	*/
			for (i=0; i < 16; i++) {
				strcat(current_line," ");
				strcat(current_line,hextable[buffer[cfrom++]]);
			}
			cfrom-=16;
			strcat(current_line,"  *");
			/*	print the bytes in ascii	*/
			ln = strlen(current_line);
			for (i=0; i < 16; i++) {
				c = buffer[cfrom++];
				if (eflag) c = ebcdic[c];
				current_line[ln++] = (c >= ' ' && c < 0x7f) ? c : '.';
			}
			current_line[ln++] = '*';
			current_line[ln] = 0;
			if (strcmp(current_line+8,previous_line+8)) {
				// Lines are different
				puts(current_line);
				strcpy(previous_line,current_line);
				dotdotdot = 0;
				++linekt;
			}
			else if (dotdotdot)
				// 
				;
			else {
				// first occurrence of a duplicate line, print ...
				puts("...");
				dotdotdot = 1;
				++linekt;
			}

		}
		tot+=numin;
	}
	while (numin == 4096 && (cfrom+tot) < end_offset);
	return 0;
}

